import click

from .processor import AssetFetcher, CreativeBuilder, zip_creatives
from ..settings import AWS_BUCKET, AWS_REGION


@click.command(help="Get the manifest from a url.")
@click.argument("url")
@click.option("-f", "--filename", default="out", help="Name of Zip File to output.")
@click.option(
    "-m",
    "--mode",
    type=click.Choice([e.name for e in AssetFetcher.Mode], case_sensitive=False),
    default=AssetFetcher.Mode.HTTP.name,
    help="Mode of Download (http or s3).",
)
@click.option("-b", "--bucket", default=AWS_BUCKET, help="Optional bucket path")
@click.option("-r", "--region", default=AWS_REGION, help="Optional region")
def build_creative(url, filename, mode, bucket, region):
    """Using the specified URL location, will retrieve, process manifest and write creatives to zip.

    Parameters:
        url (str):
            The location of the manifest to download and build image from.
        filename (str):
            The filename of the final zip file to generate.
        mode (str):
            If we are retrieving the manifest from http or aws.
        bucket (str):
            The name of the bucket where the file is stored.
        region (str):
            The region of the bucket where the file is stored.
    """
    client = AssetFetcher(AssetFetcher.Mode[mode], bucket, region)

    builder = CreativeBuilder(url, client)

    creatives = builder.build()
    click.echo("Creatives Generated!")

    zip_creatives(creatives, filename=filename)
    click.echo(f"Creatives Now Available in {filename}.zip")



from rbx.gcp.pubsub import PubSubPublisher


def notify_mdm(payload, project, topic, version):
    """Send an event to Google Pub/Sub to notify the change."""
    PubSubPublisher(project=project, topic=topic, version=version).dispatch(payload)



import googlemaps

from . import Geocode, Geotype, Location, Point, Viewport


class Client(googlemaps.Client):
    """Extend the Google Maps Client to format the response in a more usable format, which only
    includes data we care about, namely the viewport and the location type.
    """

    def geocode(self, search, country=None):
        """Force the search query to use the ISO country code as a component filter, to ensure the
        results are biased towards the country we know we are looking in.

        Returns a Geocode instance.
        """
        try:
            if country:
                response = super(Client, self).geocode(
                    search, components={"country": country}
                )
            else:
                response = super(Client, self).geocode(search)
        except googlemaps.exceptions.TransportError:
            # This exception is known to be raised when the connection is severed, which will
            # happen if the publication takes too long and the job gets killed.
            # When that happens, we just return None and move on, silencing the error.
            return None

        if not response:
            return None

        result = response[0]
        geotype = Geotype[result["geometry"]["location_type"]]
        southwest = Point(
            latitude=result["geometry"]["viewport"]["southwest"]["lat"],
            longitude=result["geometry"]["viewport"]["southwest"]["lng"],
        )
        northeast = Point(
            latitude=result["geometry"]["viewport"]["northeast"]["lat"],
            longitude=result["geometry"]["viewport"]["northeast"]["lng"],
        )
        viewport = Viewport(southwest=southwest, northeast=northeast)

        return Geocode(
            location=result.get("formatted_address", "N/A"),
            location_type=geotype,
            viewport=viewport,
        )

    def reverse_geocode(self, latlon):
        """Reverse-geocode the given latlon pair.

        The pair is provided as a (lat, lon) tuple.

        Returns a Geocode instance.
        Raise ValueError on failure.
        """
        try:
            response = super(Client, self).reverse_geocode(latlon)
        except googlemaps.exceptions.TransportError:
            # This exception is known to be raised when the connection is severed, which will
            # happen if the publication takes too long and the job gets killed.
            # When that happens, we just return None and move on, silencing the error.
            return None

        if not response:
            return None

        try:
            area = self._lookup("administrative_area_level_2", response)["long_name"]
            country = self._lookup("country", response)["short_name"]
            postal_code = self._lookup("postal_code", response)["long_name"]
            postal_town = self._lookup("postal_town", response)["long_name"]
        except ValueError as e:
            raise ValueError(f'Failed to resolve {latlon} - "{e}"')

        return Location(
            city=postal_town,
            country=country,
            formatted_address=response[0]["formatted_address"],
            place_id=response[0]["place_id"],
            postcode=postal_code,
            region=area,
        )

    def _lookup(self, field, response):
        """Loop through all addresses in the response to find the field we are after.

        When all fails, raise a ValueError.
        """
        for result in response:
            try:
                return [
                    address_component
                    for address_component in result["address_components"]
                    if field in address_component["types"]
                ][0]
            except IndexError:
                continue

        raise ValueError(field)



from os import environ

import invoke
from fabric import Connection

from .. import CONFIGURATIONS


@invoke.task
def deploy(
    ctx, host, user, key_filename, service, environment, sandboxed=None, version=None
):
    """Deploy a new Docker image and start the container on a Remote Server.

    The sandboxed flag defaults to 'false'.
    It can be overriden with the $SANDBOXED environment variable.

    The version defaults to 'develop'.
    It can be overriden with the $VERSION environment variable.

    Note that passing the parameters always takes precedence over the environment variables.
    """
    if sandboxed is None:
        sandboxed = environ.get("SANDBOXED", "false")

    if version is None:
        version = environ.get("VERSION", "latest")

    connection = Connection(
        host=host, user=user, connect_kwargs={"key_filename": key_filename}
    )

    # Login to Scoota Docker Registry
    connection.run("`aws ecr get-login --no-include-email`", echo=True)

    # Get the latest configuration for the Server we are deploying to.
    # Valid configurations are located at '/<Service>/<Env>/'
    connection.run(
        "aws s3 sync {config}/{service}/{env} {service}".format(
            config=CONFIGURATIONS, env=environment, service=service
        ),
        echo=True,
    )

    # Restart the Docker container forcing recreate so that new version is used.
    # Using `force-recreate` alone doesn't seem to be enough to use the latest version. So we
    # do `down` and then `up`.
    connection.run(
        "docker-compose -p {service}{env} -f {service}/docker-compose.yml down"
        " --volumes --remove-orphans".format(service=service, env=environment),
        echo=True,
        env={
            "SANDBOXED": sandboxed,
            "VERSION": version,
        },
        warn=True,
    )
    connection.run(
        "docker-compose -p {service}{env} -f {service}/docker-compose.yml up -d".format(
            service=service, env=environment
        ),
        echo=True,
        env={
            "SANDBOXED": sandboxed,
            "VERSION": version,
        },
    )

    # Remove all unused Docker Images
    connection.run("docker rmi `docker images -q` || true", echo=True)
    connection.run("docker rm `docker ps -a -q` || true", echo=True)


ns = invoke.Collection(deploy)



import click

from .maps import Client
from .packer import LocationPacker


@click.command(help="Fetch the Geocode for the given location.")
@click.argument("location")
@click.option("-c", "--country", help="Optional country code.")
@click.option("--key", help="google Cloud API key.", required=True)
def geocode(location, country=None, key=None):
    client = Client(key=key)
    geocode = client.geocode(location, country)

    if geocode is None:
        click.echo("Maps can't find '{}'".format(location))

    else:
        click.echo(click.style(geocode.location, fg="green"))
        click.echo(
            "{}: {} ({})".format(
                click.style("Precision", bold=True),
                geocode.location_type.name,
                geocode.location_type.value,
            )
        )
        click.echo(
            "{}: {}, {}".format(
                click.style(" > Southwest", bold=True),
                geocode.viewport.southwest[0],
                geocode.viewport.southwest[1],
            )
        )
        click.echo(
            "{}: {}, {}".format(
                click.style(" > Northeast", bold=True),
                geocode.viewport.northeast[0],
                geocode.viewport.northeast[1],
            )
        )


@click.command(help="Fetch the Location the given lat/lon.")
@click.argument("latlon")
@click.option("--key", help="google Cloud API key.", required=True)
def reverse_geocode(latlon, key=None):
    client = Client(key=key)
    location = client.reverse_geocode(latlon)

    if location is None:
        click.echo("Maps can't find '{}'".format(latlon))

    else:
        click.echo(click.style(location.formatted_address, fg="green"))
        click.echo("{}: {}".format(click.style("Country", bold=True), location.country))
        click.echo("{}: {}".format(click.style("Region", bold=True), location.region))
        click.echo("{}: {}".format(click.style("City", bold=True), location.city))
        click.echo(
            "{}: {}".format(click.style("Postcode", bold=True), location.postcode)
        )
        click.echo(
            "{}: {}".format(click.style("Place ID", bold=True), location.place_id)
        )


@click.command(help="Unpack the GeoIP databases from the DB_PATH folder.")
@click.argument("db_path")
@click.option(
    "-d",
    "--data-dir",
    default="data",
    help='Destination folder for the extracted data (defaults to "./data").',
)
@click.option(
    "-v",
    "--version",
    default="latest",
    help='GeoIP databases version (defaults to "latest").',
)
def unpack(db_path, data_dir, version):
    packer = LocationPacker(db_path=db_path, data_dir=data_dir, version=version)
    click.echo(f'Unpacking GeoIP2 databases to "{packer.data_dir}".')

    try:
        packer()
    except SystemExit:
        click.echo(click.style("Cannot find the database files.", fg="red"))
        raise SystemExit(1)
    else:
        click.echo("\U0001F44D Done.")
        click.echo(
            f'Now run "aws s3 sync {packer.data_dir} s3://<bucket>/rig/geo/{version}"'
            " to upload to S3."
        )



import logging

import click


class ColourFormatter(logging.Formatter):
    def format(self, record):
        message = super().format(record)
        colours = {
            logging.INFO: {"bold": True},
            logging.WARNING: {"fg": "yellow"},
            logging.ERROR: {"fg": "bright_red"},
            logging.CRITICAL: {"fg": "bright_white", "bg": "red"},
        }
        try:
            message = click.style(message, **colours[record.levelno])
        except KeyError:
            pass
        return message



from dataclasses import asdict, dataclass, field, InitVar
import logging
import os
from unittest import mock

import google.auth.credentials
from google.cloud.firestore import Client

from ..clients import retry
from ..exceptions import TransientException
from ..settings import RBX_PROJECT

logger = logging.getLogger("rbx.auth")


class Keystore:
    """Manages API Keys stored in Google Cloud Firestore (Native Mode)."""

    def __init__(self):
        if os.getenv("GAE_ENV", "").startswith("standard"):
            db = Client(project=RBX_PROJECT)
        else:
            credentials = mock.Mock(spec=google.auth.credentials.Credentials)
            db = Client(
                project=os.getenv("GOOGLE_CLOUD_PROJECT"), credentials=credentials
            )

        self.collection = db.collection("api_keys")

    def _key(self, document):
        """Given a DocumentSnapshot, return the fully loaded Key object."""
        return Key(**{"keystore": self, "key_id": document.id, **document.to_dict()})

    @retry.Retry(deadline=300.0)
    def create_key(
        self, email, name, campaigns=None, is_restricted=None, services=None
    ):
        """Make a new API key.

        A key is associated with a user via her email address. If a key already exists for that
        user, the key is updated instead.
        """
        key = self.get_key(email=email)
        if not key:
            key = {
                "campaigns": campaigns or [],
                "email": email,
                "is_restricted": is_restricted if is_restricted is not None else True,
                "name": name,
                "services": services if services is not None else ["*.*"],
                "status": "active",
            }

            _, ref = self.collection.add(key)
            if not ref.get().exists:
                raise TransientException(f"Failed to create key for {email}")

            key = self._key(ref.get())
        else:
            key.update(
                **{
                    "campaigns": campaigns or key.campaigns,
                    "is_restricted": is_restricted
                    if is_restricted is not None
                    else key.is_restricted,
                    "name": name or key.name,
                    "services": services or key.services,
                    "status": "active",  # the key is resuscitated
                }
            )

        return key

    def get_key(self, key_id=None, email=None):
        """Retrieve a key by ID or email."""
        if key_id:
            document = self.collection.document(key_id).get()
            if document.exists:
                return self._key(document)

        if email:
            document = next(self.collection.where("email", "==", email).stream(), False)
            if document:
                return self._key(document)

    def update_key(self, key, attributes):
        """Set the new Key values in Firestore."""
        assert type(key) is Key, f"Excepted rbx.auth.Key, got {type(key)}"
        document = self.collection.document(key.key_id)
        document.set(attributes, merge=True)


@dataclass
class Key:
    key_id: InitVar[str]
    keystore: InitVar[Keystore]
    email: str
    name: str
    campaigns: list = field(default_factory=list)
    is_restricted: bool = True
    services: list = field(default_factory=["*.*"])
    status: str = "active"

    def __post_init__(self, key_id, keystore):
        # These are defined as InitVar so that they are not part of the pickled data, and aren't
        # included in the to_dict() representation.
        self.key_id = key_id
        self.keystore = keystore

    def activate(self):
        self.update(status="active")

    def deactivate(self):
        self.update(status="inactive")

    def has_access(self, service, operation):
        """Check whether the service and operation are granted access by this Key."""
        if self.status != "active":
            return False

        for grant in self.services:
            grant = dict(zip(("service", "operation"), grant.split(".")))
            if grant["service"] == "*":
                return True
            elif grant["service"] == service and grant["operation"] == "*":
                return True
            elif grant["service"] == service and grant["operation"] == operation:
                return True

        return False

    def to_dict(self):
        return asdict(self)

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if key in self.__annotations__.keys() and key not in ("key_id", "keystore"):
                setattr(self, key, value)

        self.keystore.update_key(self, self.to_dict())


def fake_key(
    key_id,
    campaigns=None,
    email=None,
    is_restricted=True,
    name=None,
    services=None,
    status="active",
):
    return Key(
        keystore=mock.Mock(spec_set=Keystore),
        key_id=key_id,
        campaigns=campaigns or [],
        email=email or "john.doe@rip.com",
        is_restricted=is_restricted,
        name=name or "John Doe",
        services=services if services is not None else ["*.*"],
        status=status,
    )



from invoke import Config, Program

from .tasks import deploy


class BuildtoolsConfig(Config):
    prefix = "rbx"


program = Program(config_class=BuildtoolsConfig, namespace=deploy.ns, version="3.0.2")



import json
import operator
import zipfile
from collections import namedtuple
from collections.abc import MutableSequence
from enum import Enum, auto
from io import BytesIO

import cv2
import numpy as np
import requests

from .utils import flatten, make_transparent_image, make_white_image, px_to_int
from ..aws.s3 import S3Bucket
from ..exceptions import FatalException

CreativeVariant = namedtuple("CreativeVariant", ["filename", "name", "image"])
Variant = namedtuple("Variant", ["name", "multiplier"])


class FetchMode(Enum):
    S3 = auto()
    HTTP = auto()


class Modal(type):
    def __call__(cls, mode, *args, **kwargs):
        if mode is FetchMode.S3:
            return type.__call__(S3AssetFetcher, *args, **kwargs)
        else:
            return type.__call__(HttpAssetFetcher)


class AssetFetcher(metaclass=Modal):
    Mode = FetchMode

    def get_image(self, *args, **kwargs):
        return NotImplementedError

    def get_manifest(self, *args, **kwargs):
        return NotImplementedError


class HttpAssetFetcher(AssetFetcher):
    def __init__(self):
        self.base_url = None

    def get_image(self, url, base_url=None):
        """Retrieve Image from http, decode it using opencv and return the decoded image.

        Parameters:
            url (str):
                The location of the image to retrieve.
            base_url (str):
                The base location of the image to retrieve, if overriding the manifest.

        Returns:
            ndarray:
                An opencv representation of the retrieved image,
        """
        if self.base_url is None and base_url is None:
            raise FatalException(f"A Base URL must be provided: {url}")

        else:
            base_url = base_url if base_url is not None else self.base_url
            response = requests.get(f"{base_url}{url}")
            if not response.status_code == 200:
                raise FatalException(
                    f"Image not found at specified location: {base_url}{url}"
                )
            img_data = response.content
        decoded_image = cv2.imdecode(
            np.asarray(bytearray(img_data)), cv2.IMREAD_UNCHANGED
        )
        # if the image fails to be read, then it will be None
        if decoded_image is None:
            raise FatalException(f"Image has an invalid format: {base_url}{url}")
        return decoded_image

    def get_manifest(self, url):
        """Retrieve Manifest from HTTP (expecting a JSON format) and return the decoded dict.

        Parameters:
            url (str):
                The location of the manifest to retrieve.

        Returns:
            dict:
                The decoded JSON, stored as a dict.
        """
        response = requests.get(url)
        if response.status_code != 200:
            raise FatalException(f"Manifest not found at specified location: {url}")
        try:
            manifest = response.json()
        except json.decoder.JSONDecodeError:
            raise FatalException(f"Manifest has an invalid format: {url}")

        self.base_url = manifest.get("base_url")
        if self.base_url.startswith("//"):
            self.base_url = f"http:{self.base_url}"
        return manifest


class S3AssetFetcher(AssetFetcher):
    def __init__(self, bucket=None, region=None, endpoint_url=None):
        """
        Setup the asset fetcher and prepare a client.

        Parameters:
            bucket (str):
                The bucket to use when retrieving from S3.
            region (str):
                The region to use when retrieving from S3.
            endpoint_url (str):
                Override the endpoint. Only used for testing and should normally not be set.
        """
        self.client = S3Bucket(name=bucket, region=region, endpoint_url=endpoint_url)

    def get_image(self, url):
        """Retrieve Image from S3, decode it using opencv and return the decoded image.

        Parameters:
            url (str):
                The location of the image to retrieve.

        Returns:
            ndarray:
                An opencv representation of the retrieved image,
        """
        image = self.client.get_key(f"{url}")
        if image is None:
            raise FatalException(
                f"Image not found at specified location: " f"{self.client.name}/{url}"
            )
        img_data = image.get()["Body"].read()

        decoded_image = cv2.imdecode(
            np.asarray(bytearray(img_data)), cv2.IMREAD_UNCHANGED
        )
        # if the image fails to be read, then it will be None
        if decoded_image is None:
            raise FatalException(
                "Image has an invalid format: {}{}".format(
                    self.client.name + "/",
                    url,
                ),
            )
        return decoded_image

    def get_manifest(self, url):
        manifest_data = self.client.get_key(url)
        if manifest_data is None:
            raise FatalException(
                f"Manifest not found at specified location: {self.client.name}/{url}"
            )
        try:
            file_content = manifest_data.get()["Body"].read()
            manifest = json.loads(file_content)
        except UnicodeDecodeError:
            raise FatalException(f"Manifest has an invalid format: {url}")

        return manifest


class ComponentSequence(MutableSequence):
    """A mutable sequence of components."""

    def __init__(self):
        self.components = []

    def __len__(self):
        """The length includes the length of its components."""
        length = len(self.components)
        for component in self.components:
            length += len(component)
        return length

    def __getitem__(self, key):
        return self.components[key]

    def __setitem__(self, key, value):
        self.components[key] = value

    def __delitem__(self, key):
        del self.components[key]

    def append(self, component, client, variant):
        result = make_component(component, client, variant)
        self.components.append(result)

    def insert(self, value):
        self.components.insert(value)


class CreativeBuilder:
    """Builds and returns all variations of Creatives that the manifest describes.

    As there can potentially be many variations of creatives, the creative builder is responsible
    for calculating all variations, passing one variant to each Creative for rendering.
    This allows us to calculate and apply the variant logic in a central location.
    """

    def __init__(self, manifest_url, client):
        self.manifest = client.get_manifest(manifest_url)
        self.client = client
        self.variations = self.unpack_variants()

    def unpack_variants(self):
        """Calculates all variations that need to be generated for a given manifest.

        Returns:
            list:
                A list of Variant objects that describe the changes that need to be applied to the
                creative.
        """
        variations = []

        pixel_ratios = self.manifest.get("pixel_ratios")

        for pixel_multiplier, pixel_name in pixel_ratios.items():
            variations.append(
                Variant(name=pixel_name, multiplier=int(pixel_multiplier))
            )
        return variations

    def build(self):
        """Builds and returns all possible variations of Creatives, as described in the Manifest.

        Returns:
            list:
                A list of CreativeVariant objects that describe the filename, human readable name
                and image data.
        """
        creatives = []

        for variant in self.variations:
            creative = Creative(
                manifest=self.manifest, client=self.client, variant=variant
            )

            creatives.append(
                CreativeVariant(
                    filename=f"final_image{variant.name}.png",
                    name=creative.name,
                    image=creative.render(),
                )
            )

        return creatives


class Creative:
    def __init__(self, client, manifest, variant):
        self.manifest = manifest
        self.client = client
        self.variant = variant

        root = self.manifest.get("root_component")
        self.height = px_to_int(root["css"]["height"]) * variant.multiplier
        self.width = px_to_int(root["css"]["width"]) * variant.multiplier
        self.left = px_to_int(root["css"]["left"]) * variant.multiplier
        self.top = px_to_int(root["css"]["top"]) * variant.multiplier

        self.root_component = make_component(root, self.client, variant)

    @property
    def name(self):
        """Uses the creative variant details to return the appropriate human readable name.

        Returns:
            str:
                The human readable name.
        """
        name = "Creative"

        if self.variant.multiplier > 1:
            name = f"Retina {name} {self.variant.multiplier}x Ratio"

        return name

    def render(self):
        """Builds an opencv image representation of the manifest and return.

        We start by making a white canvas, whose size is defined by the root component. We then
        call the compile function of the root component, which should call the nested compile/render
        functions of all sub components. This is then returned as a nested, ordered list of images
        that are ready to be flattened and applied to the solid root image (base canvas).

        Returns:
            ndarray:
                An opencv representation of the final image.
        """

        root_image = make_white_image(height=self.height, width=self.width)

        ordered_images = self.root_component.compile()

        flattened_images = list(flatten(ordered_images))
        for entry in flattened_images:
            root_image = merge_images(
                background_image=root_image,
                overlay_image=entry,
                left=self.left,
                top=self.top,
                transparent_background=False,
            )

        return root_image


class Component(ComponentSequence):
    def __init__(self, component, client, variant):
        super(Component, self).__init__()
        self.client = client
        self.variant = variant
        self.top = px_to_int(component["css"]["top"]) * variant.multiplier
        self.left = px_to_int(component["css"]["left"]) * variant.multiplier
        self.height = px_to_int(component["css"]["height"]) * variant.multiplier
        self.width = px_to_int(component["css"]["width"]) * variant.multiplier
        self.z_index = component["css"]["z-index"]


class LayoutComponent(Component):
    """A Component which has child Component."""

    def __init__(self, component, client, variant):
        """Initialises the ImageComponent.

        For a LayoutComponent, we need to add the child components that it has underneath it.

        Parameters:
            component (dict): The details of the component.
            client (S3Bucket): The S3 client to use instead of http, if needed.
            variant (Variant): The variant details to apply to the component.
        """
        super(LayoutComponent, self).__init__(
            component=component, client=client, variant=variant
        )
        children = component.get("component_manifests")

        for child_component in children:
            self.append(child_component, self.client, self.variant)

    def compile(self):
        """
        Gets all sub-components of the layer and applies each of them to a transparent canvas.

        This allows the positional data of both the image and the layer to be correctly applied, for
        an accurate final merge with the base canvas.

        Returns:
            list:
                A list of ndarray images, each applied to the transparent canvas.
        """
        transparent_image = make_transparent_image(self.height, self.width)

        images = []

        for entry in sorted(self.components, key=operator.attrgetter("z_index")):
            flattened_entry = flatten(entry.compile())
            for flat_entry in flattened_entry:
                images.append(
                    merge_images(
                        background_image=transparent_image,
                        overlay_image=flat_entry,
                        left=entry.left,
                        top=entry.top,
                        transparent_background=True,
                    )
                )

        return images


class ImageComponent(Component):
    """A component that describes an image."""

    def __init__(self, component, client, variant):
        """Initialises the ImageComponent, storing the location of the Image file.

        Parameters:
            component (dict): The details of the component.
            client (S3Bucket): The S3 client to use instead of http, if needed.
            variant (Variant): The variant details to apply to the component.
        """
        super(ImageComponent, self).__init__(
            component=component, client=client, variant=variant
        )
        self.url = component.get("url")
        self.ext = component.get("ext")

    def compile(self):
        """Return the rendered image in an iterable format.

        Returns:
            list:
                Returns the image in an iterable, for easier handling in the base render function.
        """
        return [self.render()]

    def render(self):
        """Retrieve Image from http or S3, decode it using opencv and return the decoded image.

        Returns:
            ndarray:
                An opencv representation of the retrieved image,
        """
        return self.client.get_image(f"{self.url}{self.variant.name}{self.ext}")


def generate_name(multiplier):
    """Takes a retina multiplier and returns the appropriate human readable name.

    Parameters:
        multiplier (int): The retina multiplier associated with the image.

    Returns:
        str:
            The human readable name.
    """
    name = "Creative"

    if multiplier > 1:
        name = f"Retina {name} {multiplier}x Ratio"

    return name


def make_component(component, client, variant):
    """Takes a component and returns the appropriate sub-component.

    Parameters:
        component (dict): The details of the component.
        client (S3Bucket): The S3 client to use instead of http, if needed.
        variant (Variant): The variant details to apply to the component.

    Returns:
        any:
            A instance of a component sub-type.
    """
    component_type = component.get("component_type")

    if component_type == "layout":
        return LayoutComponent(component, client, variant)

    elif component_type == "image":
        return ImageComponent(component, client, variant)


def merge_images(
    background_image, overlay_image, left, top, transparent_background=False
):
    """Puts the overlay image on top of the background image.

    Will use the left and top to calculate the position of the overlay image relative to the
    background image. Any transparency the top image has will be applied, showing the background
    image where applicable. If transparent_background is set, the overlay image will completely
    overwrite the underlying image.

    This logic was sourced from https://stackoverflow.com/questions/14063070/14102014#14102014 with
    minor modifications to fit our needs.

    Parameters:
        background_image (ndarray):
            The opencv representation of the background image, on which the new image will apply.
        overlay_image (ndarray):
            The opencv representation of the overlay image, which is placed over background image.
        left (int):
            How far, in pixels, from the left the overlay image should be applied.
        top (int):
            How far, in pixels, from the top the overlay image should be applied.
        transparent_background (boolean):
            If the background is transparent, and therefore should be overwritten.

    Returns:
        ndarray:
            An opencv representation of the combined image.
    """
    # we need to copy the initial background/overlay params otherwise the originals get mutated
    background_copy = background_image.copy()
    overlay_copy = overlay_image.copy()

    # if the background doesn't have an alpha layer, add one (will always be solid)
    if background_copy.shape[-1] < 4:
        b_channel, g_channel, r_channel = cv2.split(background_copy)
        alpha_channel = np.ones(b_channel.shape, dtype=b_channel.dtype) * 255
        background_copy = cv2.merge((b_channel, g_channel, r_channel, alpha_channel))
    # if the overlay doesn't have an alpha layer, add one (will always be solid)
    if overlay_copy.shape[-1] < 4:
        b_channel, g_channel, r_channel = cv2.split(overlay_copy)
        alpha_channel = np.ones(b_channel.shape, dtype=b_channel.dtype) * 255
        overlay_copy = cv2.merge((b_channel, g_channel, r_channel, alpha_channel))

    # reposition overlay image
    y1, y2 = top, top + overlay_copy.shape[0]
    x1, x2 = left, left + overlay_copy.shape[1]

    # if the background is transparent, overwrite all pixels in the background with the overlay
    if transparent_background:
        background_copy[y1:y2, x1:x2] = overlay_copy
    else:
        # apply overlay image to solid background image, keeping the background solid and blending
        # the colour and transparency of the overlay with the background colour to create an
        # accurate combined image

        # blend alpha channels
        alpha_s = overlay_copy[:, :, 3] / 255.0
        alpha_l = 1.0 - alpha_s

        # merge colours and transparency for final effect
        for c in range(0, 3):
            background_copy[y1:y2, x1:x2, c] = (
                alpha_s * overlay_copy[:, :, c]
                + alpha_l * background_copy[y1:y2, x1:x2, c]
            )
    return background_copy


def zip_creatives(images, filename=None):
    """Combine all creatives in the iterable to a ZIP format.

    Parameters:
        images (list):
            A list of CreativeVariants with the filename, human name and ndarray image data.
        filename (str):
            The name of the zip file to output. If None, then returns a buffer representation.

    Returns:
        any:
            Returns either None if writing to file or BytesIO buffer if not saving file.
    """
    buffer = None
    if filename is None:
        buffer = BytesIO()
        zipf = zipfile.ZipFile(buffer, "w")
    else:
        zipf = zipfile.ZipFile(f"{filename}.zip", "w", zipfile.ZIP_DEFLATED)

    for data in images:
        retval, buf = cv2.imencode(".png", data.image)
        zipf.writestr(data.filename, buf)

    zipf.close()
    return buffer



import invoke
from invoke import Collection

from . import image, misc, pypi


@invoke.task
def clean(ctx):
    """
    Remove all intermediate files from the source tree.
    """
    ctx.run("git clean -xdf")


ns = Collection()
ns.add_task(clean)
ns.add_collection(pypi)
ns.add_collection(image)
ns.add_collection(misc)



from collections import namedtuple
from enum import Enum


Geocode = namedtuple("Geocode", ["location", "location_type", "viewport"])
Location = namedtuple(
    "Location",
    ["city", "country", "formatted_address", "place_id", "postcode", "region"],
)
Point = namedtuple("Point", ["latitude", "longitude"])


class Viewport(namedtuple("Viewport", ["southwest", "northeast"])):
    def __contains__(self, point):
        """A Point is in the Viewport when its lat/long coordinates are within the SW/NE bounding
        box.

        Latitude
               ^
               |              NE
               +-----+-------+
               |     |       |
               x--------P    |
               |     |  |    |
               +-----+--|----+
               |   SW|  |    |
               o-----+--x----+----------> Longitude

        """
        if type(point) is Point:
            return all(
                [
                    point.latitude >= self.southwest.latitude,
                    point.latitude <= self.northeast.latitude,
                    point.longitude >= self.southwest.longitude,
                    point.longitude <= self.northeast.longitude,
                ]
            )

        return False


class Geotype(Enum):
    ROOFTOP = (
        "Indicates that the returned result is a precise geocode for which we have location "
        "information accurate down to street address precision."
    )
    RANGE_INTERPOLATED = (
        "Indicates that the returned result reflects an approximation (usually on a road) "
        "interpolated between two precise points (such as intersections). Interpolated results "
        "are generally returned when rooftop geocodes are unavailable for a street address."
    )
    GEOMETRIC_CENTER = (
        "Indicates that the returned result is the geometric center of a result such as a "
        "polyline (for example, a street) or polygon (region)."
    )
    APPROXIMATE = "Indicates that the returned result is approximate."
    MANUAL = "Indicates that the Coordinates were added manually."
    MISSING = "Indicates that a result would not be found."
    MULTIPLE = "Indicates that the Coordinates are an alternative location for an existing location."



from os import path

import invoke
from six import PY2


def run(ctx, command):
    result = ctx.run(command, hide="out").stdout
    if PY2:
        result = result.decode("utf-8")
    return result.rstrip()


@invoke.task(name="git-next-tag")
def get_next_tag(ctx):
    """A helper to generate the next tag based on the current git tag.

    To use in Jenkins job:

        >>> export VERSION=`invoke misc.git-next-tag`

    Expects an integer, otherwise will start at 1.
    """
    tag = run(ctx, "git describe --tags $(git rev-list --tags --max-count=1)")
    try:
        tag = int(tag)
    except ValueError:
        tag = 0
    print(tag + 1)


@invoke.task(name="git-merge-desc")
def get_merge_desc(ctx):
    """A helper to get the description of a merge commit.

    To use in a Jenkins job:

        >>> export DESCRIPTION=`invoke misc.git-merge-desc`

    When merging a PR the subject line will be added to the body of the merge
    commit.
    """
    print(run(ctx, 'git log -1 --pretty=format:"%b"'))


@invoke.task(name="npm-version")
def get_npm_version(ctx):
    """A helper to get the npm package version from a package.json file.

    To use in Jenkins job:

        >>> export VERSION=`invoke misc.npm-version`

    """
    if path.exists("package.json"):
        print(run(ctx, "cat package.json | jq .version -r"))
    else:
        print("unknown")



"""
Helper module to represent scheduled jobs.
"""


class JobSpec:
    """
    Represents a job spec.

    This is a Value Object, and doesn't convey any logic.
    All this does is some very limited field checking.
    """

    fields = ("task", "args", "kwargs", "schedule", "timeout")

    def __init__(self, name, fields):
        if not all([rv in fields.keys() for rv in ("task", "schedule")]):
            raise AttributeError("These fields are required: {}".format(self.fields))

        job_spec = dict([(k, v) for k, v in fields.items() if k in self.fields])
        if "args" not in job_spec.keys():
            job_spec["args"] = ()
        if "kwargs" not in job_spec.keys():
            job_spec["kwargs"] = {}
        if "timeout" not in job_spec.keys():
            job_spec["timeout"] = None

        self.__dict__ = job_spec

    def __str__(self):
        return "{schedule} {task}".format(**self.__dict__)

    def __repr__(self):
        return "<JobSpec [{}]>".format(str(self))


class Jobs(dict):
    """Iterable of JobSpecs."""

    def __setitem__(self, key, value):
        spec = JobSpec(key, value)
        super().__setitem__(key, spec)

    @staticmethod
    def from_dict(jobs_as_dict):
        """
        Constructor that creates a Jobs instance from a dictionary.
        The dict is expected to have be valid and have the following format:

        >>> {
        >>>     'job_name': {
        >>>         'task': 'full.path.to.method',
        >>>         'args': (1, 'arg2'),
        >>>         'kwargs': {'kw1': 'value1'}
        >>>         'schedule': 'CRON expression',
        >>>         'timeout': 180,
        >>>     },
        >>>     ...
        >>> }

        """
        jobs = Jobs()
        for key, value in jobs_as_dict.items():
            jobs[key] = value
        return jobs



import logging

from . import Client, HttpAuth

logger = logging.getLogger(__name__)


class ReportingClient(Client):
    """API Client for the Reporting Service."""

    def __init__(self, endpoint, token):
        super().__init__()
        self.ENDPOINT = endpoint.rstrip("/")
        self.token = token

    @property
    def auth(self):
        """The Reporting Service uses Digest Authentication."""
        return HttpAuth(self.token, key="X-RBX-TOKEN")

    def create_schedule(self, body, campaign, cron, format, name, recipients, reports):
        """Create a scheduled job.

        Return the job_id of the job just created.
        """
        response = self._post(
            "/schedule",
            data={
                "body": body,
                "campaign": campaign,
                "cron": cron,
                "format": format,
                "name": name,
                "recipients": recipients,
                "reports": reports,
            },
        )
        return response.get("job_id")

    def delete_schedule(self, job_id):
        """Create the scheduled job."""
        response = self._post("/schedule/delete", data={"job_id": job_id})
        return response.get("job_id")

    def pause_schedule(self, job_id):
        """Pause the scheduled job."""
        response = self._post("/schedule/pause", data={"job_id": job_id})
        return response.get("job_id")

    def resume_schedule(self, job_id):
        """Resume the scheduled job."""
        response = self._post("/schedule/resume", data={"job_id": job_id})
        return response.get("job_id")

    def update_schedule(
        self, body, campaign, cron, format, job_id, name, recipients, reports
    ):
        """Update the scheduled job."""
        response = self._post(
            "/schedule",
            data={
                "body": body,
                "campaign": campaign,
                "cron": cron,
                "format": format,
                "job_id": job_id,
                "name": name,
                "recipients": recipients,
                "reports": reports,
            },
        )
        return response.get("job_id")

    def get_report(
        self,
        breakdowns,
        metrics=None,
        custom_metrics=None,
        campaign=None,
        anomalies=None,
        filters=None,
        date_range=None,
        start_date=None,
        end_date=None,
        format=None,
        step=None,
        start=None,
        order_by=None,
        time_series=None,
        timezone=None,
    ):
        """Retrieve report data using the specified parameters.

        Parameters:
            breakdowns (list):
                The list of requested breakdowns.
            metrics (list):
                The list of requested metrics.
            custom_metrics (dict):
                A keyed object detailing the labels and units of any custom metrics.
                The key must be the position of the custom metric in the `metrics` parameter,
                beginning with `0`, as a string.

                By default, custom metrics are labelled `N/A` and use the `number` unit.
            campaign (int):
                The id of the Campaign.
            anomalies (str):
                The report anomalies flag is used to determine which variant of the metric to
                 return: exluding anomalies, including anomalies, or anomalies only.

                Valid values for this flag are: `excluded`, `included`, and `only`.
            filters (list[breakdown,operator,value]):
                Filter results on the given Breakdown having the value matching the given operator.
                Supported operators are: `eq`, `neq`, `in`, `nin`, and `like`
            date_range (string):
                A standard date range.
                One of `up_to_yesterday`, `campaign_life`, `today`, `yesterday`, `last_N_days`,
                `last_N_weeks`, `last_full_week`, or `last_full_month`.

                If omitted, all data up to yesterday midnight returned.
            start_date (string):
                The start date and time in ISO 8601 format.

                If omitted, the start date defaults to the oldest date on record.
                For a single campaign, that would be the first time the campaign received an event.
            end_date (string):
                The end date and time in ISO 8601 format.

                If omitted, the end date defaults to yesterday midnight.
            format (string):
                Specify what format the results should be returned in.
                Available formats are `csv`, `chartjs`, `tabular`, and `xlsx`.

                By default results are returned in JSON `tabular` format.
            step (int):
                The number of results to return. Use to limit the number of rows.

                By default all rows are returned.
            start (int):
                Use alongside `step` to fetch a specific set of results.
                `start` should be the number of the expected first result on the requested page.

                e.g.: using a `step` of 24, to get page 2, use `step=24, start=25`
            start (int):
                Use alongside `step` to fetch a specific set of results.
                `start` should be the number of the expected first result on the requested page.

                e.g.: using a `step` of 24, to get page 2, use `step=24, start=25`
            order_by (list[item,order]):
                Specify how the response should be ordered.

                The `order_by` parameter is expected to be a list of `item: sort` dictionary,
                where `item` is any Breakdown or Metric, and `sort` is either `ASC` or `DESC`.
            time_series (bool):
                Whether to return a time series.
            timezone (str):
                The time zone name of the request,
                If omitted, the time zone is assumed to be UTC.

        Returns:
            Dict.
            A dict response containing the "headers" and "rows", with each row ordered to match the
            order of the headers.
        """
        kwargs = locals()
        data = {}

        for label, value in kwargs.items():
            if label != "self" and value is not None:
                data[label] = value

        return self._post("/report", data=data)



from collections import namedtuple
from datetime import datetime
from decimal import Decimal
import json
import logging
from typing import NamedTuple

import arrow

from . import Client, ClientError, retry, TransientServerError

logger = logging.getLogger(__name__)

decimal_metrics = ["actual_pub_payout", "dooh_impmultiply"]


class EntrySet(NamedTuple):
    timestamp: datetime
    entries: list


class SlicerClient(Client):
    """HTTP client for the u-Slicer API.

    https://confluence.iponweb.net/display/UPLATFORMKB/API

    >>> from rbx.clients.uslicer import SlicerClient
    >>> client = SlicerClient(token='eyJhbGciOiJFU...')

    """

    TIME_FORMAT = "YYYY-MM-DD"
    DEFAULT_TIMEOUT = 60
    ENDPOINT = "https://uslicer.iponweb.com/API/v2/query"

    def __init__(self, token, project_name="rockabox", slicer_name="Traffic"):
        super().__init__()
        self.token = token
        self.project_name = project_name
        self.slicer_name = slicer_name

        self.breakdowns = []
        self.metrics = []

    @property
    def entry(self):
        assert (
            self.breakdowns and self.metrics
        ), "An entry only exists after a query has been executed"
        return namedtuple(
            "Entry", map(lambda x: x.split(".")[-1], self.breakdowns + self.metrics)
        )

    def fetch(self, breakdowns, metrics, start_date, end_date, filters=None):
        """fetch u-Slicer metrics in hourly batches.

        The method uses the same signature as the `query()` method, but instead returns an
        iterator, which yields hourly results.

        Unlike `query()`, however, this methods allows fetching data with more precise start and
        end dates, using hours as well as days.

        Note that both the start and end dates are rounded to the previous hour. As uSlicer does
        not handle that level of precision, this is made cleared that way. The `timestamp` datetime
        attribute of each iteration will use that rounded date.
        """
        start_date = arrow.get(start_date).floor("hour")
        end_date = arrow.get(end_date).floor("hour")
        filters = filters or []
        for r in arrow.Arrow.range("hour", arrow.get(start_date), arrow.get(end_date)):
            results = self.query(
                breakdowns,
                metrics,
                start_date=r.format(self.TIME_FORMAT),
                end_date=r.format(self.TIME_FORMAT),
                filters=filters + [("hour", "equals", [r.format("H")])],
            )

            yield EntrySet(timestamp=r.datetime, entries=results)

    @retry.Retry(deadline=300.0)
    def query(
        self,
        breakdowns,
        metrics,
        start_date,
        end_date,
        filters=None,
        none_default="N/A",
    ):
        """Fetch u-Slicer metrics.

        The metrics are returned for the given date range. The start and end dates are given as
        datetime objects.

        The optional `filters` argument takes a LIST of filter tuples.
        e.g.:
        >>> filters = [(Breakdown.CAMPAIGN, Operator.EQUALS, [123234453])]

        The values are returned as a list of Entry objects.
        """
        self.breakdowns = breakdowns
        self.metrics = metrics

        params = {
            "project_name": self.project_name,
            "slicer_name": self.slicer_name,
            "token": self.token,
            "data_fields": metrics,
            "end_date": arrow.get(end_date).format(self.TIME_FORMAT),
            "limit": -1,
            "split_by": breakdowns,
            "start_date": arrow.get(start_date).format(self.TIME_FORMAT),
            "timezone": 0,
        }

        # Resolve each filter tuple in the filters list.
        _filters = []
        if filters:
            for breakdown, operator, value in filters:
                _filters.append({"name": breakdown, "match": operator, "value": value})

        if _filters:
            params["filters"] = _filters

        logger.debug(f"Query params: {params}")
        try:
            response = self.request("post", self.ENDPOINT, data=params)
        except ClientError as exc:
            try:
                message = json.loads(str(exc))
            except json.decoder.JSONDecodeError as e:
                raise exc from e
            else:
                if message.get("status") == "internal_error":
                    raise TransientServerError(
                        exc,
                        details=exc.details,
                        status_code=exc.status_code,
                        url=exc.url,
                    )
                else:
                    raise

        results = []

        for row in response["rows"]:

            # Annoyingly Iponweb have decided to call the Panel ID field 'supply.panel_id',
            # which is not namedtuple-friendly. So here we apply that logic to all fields in an
            # attempt to be future-proof, namely to remove all dot prefixes.
            # Plus, empty/null breakdowns get replaced with custom value (which defaults to 'N/A').
            fields = {}
            for i, breakdown in enumerate(breakdowns):
                fields[breakdown.split(".")[-1]] = row["name"][i] or none_default

            fields.update(
                dict(
                    [
                        (
                            metric["name"],
                            Decimal(str(metric["value"]))
                            if metric["name"] in decimal_metrics
                            else metric["value"],
                        )
                        for metric in row["data"]
                    ]
                )
            )

            results.append(self.entry(**fields))

        logger.debug(f"Found {len(results)} entries.")

        return results



from os import environ, getcwd, path

import invoke

from .. import REGISTRY

BASEPATH = path.basename(getcwd())


@invoke.task(name="build", iterable=["build_arg"])
def build_image(ctx, build_arg=None, name=None, version=None, target=None, test=False):
    """Build the docker image.

    The image name defaults to the name of the working folder.
    It can be overriden with the $SERVICE environment variable.

    The version defaults to 'develop'.
    It can be overriden with the $VERSION environment variable.

    Note that parameters always take precedence over environment variables.
    """
    command = ["docker build"]

    if target:
        command.append("--target {}".format(target))

    if test:
        command.append("--build-arg INSTALL_TEST_DEPENDENCIES=yes")

    for arg in build_arg:
        command.append("--build-arg {}".format(arg))

    command.append(
        "-t {registry}/{repository}:{version} .".format(
            registry=REGISTRY,
            repository=name or environ.get("SERVICE", BASEPATH),
            version=version or environ.get("VERSION", "develop"),
        )
    )

    command = " ".join(command)
    ctx.run(command)


@invoke.task(name="upload")
def upload_image(ctx, name=None, version=None):
    """Login then upload the docker image to the registry.

    The image name defaults to the name of the working folder.
    It can be overriden with the $SERVICE environment variable.

    The version defaults to 'develop'.
    It can be overriden with the $VERSION environment variable.

    Needless to say, this image must have been already built.

    Note that parameters always take precedence over environment variables.
    """
    ctx.run("`aws ecr get-login --no-include-email`")
    command = "docker push {registry}/{repository}:{version}".format(
        registry=REGISTRY,
        repository=name or environ.get("SERVICE", BASEPATH),
        version=version or environ.get("VERSION", "develop"),
    )
    ctx.run(command)



import logging
import os

import boto3
import botocore

from rbx.exceptions import AWSException

logger = logging.getLogger(__name__)


class S3Bucket:
    """An AWS S3 Bucket.

    Wraps the boto3.Bucket class to provide a higher level of abstraction, as needed by the
    Platform.

    The main purpose of the create() and delete() method is to aid integration testing.
    We therefore skip any CORS or permission setting when creating buckets.
    In a production environment, these buckets are expected to already exist, and to have been
    configured accordingly.

    The only scenario where endpoint_url is set is in testing and development. In a normal
    setting, this should not be set.

    """

    def __init__(self, name, region, endpoint_url=None):
        """The name may include the bucket prefix.

        e.g.: bucket-name/path/to/dir/

        In which case the prefix instance attribute will be automatically prefixed to all keys.
        """
        self.name, _, self.prefix = name.partition("/")
        if self.prefix and not self.prefix.endswith("/"):
            self.prefix += "/"
        self.region = region

        self.resource = boto3.resource("s3", endpoint_url=endpoint_url)
        self.bucket = self.resource.Bucket(self.name)

        self.create()

    def clone(self, source, dest):
        """Clone the given S3 key to the location given in 'dest'."""
        source_key = self.prefix + source
        dest_key = self.prefix + dest
        try:
            for key in self.bucket.objects.filter(Prefix=source_key):
                d = dest_key + key.key[len(source_key) :]
                copy_source = {"Bucket": self.name, "Key": key.key}
                self.bucket.copy(copy_source, d)

        except botocore.exceptions.ClientError as e:
            logger.error(e, extra=e.response)
            raise AWSException(e)

    def copy(self, filename, key, content_type="binary/octet-stream", download=False):
        """Copy a file to the given S3 key.

        The filename will be read, so filename is expected to be the path to the file itself.

        When download is True,  we add the "ContentDisposition: attachment" header so that it makes
        browsers download the file when accessed directly.
        """
        content_type = (
            content_type or "binary/octet-stream"
        )  # Content-Type cannot be None
        options = {"ContentType": content_type}
        if download:
            options["ContentDisposition"] = "attachment"

        try:
            self.bucket.upload_file(filename, self.prefix + key, ExtraArgs=options)
        except botocore.exceptions.ClientError as e:
            logger.error(e, extra=e.response)
            raise AWSException(e)

    def create(self):
        """Create the S3 Bucket.

        Only create the bucket if it does not already exist.
        """
        if not self.exists():
            self.bucket = self.resource.create_bucket(
                Bucket=self.name,
                CreateBucketConfiguration={"LocationConstraint": self.region},
            )

    def delete(self):
        """Delete the S3 Bucket."""
        if self.exists():
            for key in self.bucket.objects.all():
                key.delete()

            self.bucket.delete()

    def delete_keys(self, path, startswith=None):
        """Delete keys from the S3 Bucket

        Will delete all keys in the Bucket given a key unless you specify a prefix, then it will
        only delete files which start with that particular prefix.
        """
        path_prefix = self.prefix + path
        deleted_keys = []

        try:
            for key in self.bucket.objects.filter(Prefix=path_prefix):
                if not startswith:
                    key.delete()
                    deleted_keys.append(key)
                elif key.key.startswith(os.path.join(path_prefix, startswith)):
                    key.delete()
                    deleted_keys.append(key)
            return deleted_keys
        except botocore.exceptions.ClientError as e:
            logger.error(e, extra=e.response)
            raise AWSException(e)

    def exists(self):
        """Check whether the Bucket exists in S3."""
        try:
            self.resource.meta.client.head_bucket(Bucket=self.name)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = e.response["Error"]["Code"]
            if error_code == "404":
                return False
            else:
                logger.error(e)

        return True

    def get_key(self, key):
        """Check whether the given key exists in the bucket.
        Will return the object or None.
        """
        object = self.bucket.Object(self.prefix + key)

        try:
            object.load()
        except botocore.exceptions.ClientError as e:
            error_code = e.response["Error"]["Code"]
            if error_code == "404":
                return
            else:
                logger.error(e)
        else:
            return object

    def save(self, fileobj, key, content_type="binary/octet-stream"):
        """Save a file-like object to the given S3 key."""
        options = {"ContentType": content_type}

        try:
            self.bucket.upload_fileobj(fileobj, self.prefix + key, ExtraArgs=options)
        except ValueError as e:
            raise AWSException(e)
        except botocore.exceptions.ClientError as e:
            logger.error(e, extra=e.response)
            raise AWSException(e)



# Copyright 2022 Rockabox Media Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup

AUTH = ["google-cloud-firestore>=2.3,<3"]
MANIFEST = ["boto3<2", "opencv-python-headless==4.4.0.46"]
MAPS = [
    "googlemaps>=4.4.2,<5",
]
NOTIFICATIONS = [
    "google-cloud-pubsub>=2.9,<3",
]
QUEUES = [
    "redis>3.5,<4.2",
    "hiredis==2.0.0",
    "rq==1.10.1",
    "rq-scheduler==0.11.0",
]
STORAGE = ["google-cloud-storage>=2.1,<3"]
TASKS = ["google-cloud-tasks>=2.7,<3"]


setup(
    name="rbx",
    version="3.0.2",
    license="Apache 2.0",
    description="Scoota Platform utilities",
    long_description="A collection of common tools for Scoota services.",
    url="http://scoota.com/",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Internet",
    ],
    author="The Scoota Engineering Team",
    author_email="engineering@scoota.com",
    python_requires=">=3.7",
    install_requires=[
        "arrow>=1,<2",
        "cachetools>=5,<6",
        "Click<9",
        "colorama",
        "PyYAML>=5.4.1",
        "requests>=1.21.1",
    ],
    extras_require={
        # These are requirement bundles required for specific feature sets.
        "auth": AUTH,
        "buildtools": [
            "bumpversion==0.5.3",
            "check-manifest",
            "fabric~=2.5.0",
            "twine",
        ],
        "geo": MAPS,
        "manifest": MANIFEST,
        "notifications": NOTIFICATIONS,
        "platform": AUTH + MANIFEST + MAPS + NOTIFICATIONS + QUEUES,
        "queues": QUEUES,
        "storage": STORAGE,
        "tasks": TASKS,
        # Include them all for the test suite.
        "test": AUTH + MANIFEST + MAPS + NOTIFICATIONS + QUEUES + STORAGE + TASKS,
    },
    entry_points={
        "console_scripts": [
            "buildtools = rbx.buildtools.cli:program.run [buildtools]",
            "buildcreative = rbx.manifest.cli:build_creative [manifest]",
            "geocode = rbx.geo.cli:geocode [geo]",
            "reverse_geocode = rbx.geo.cli:reverse_geocode [geo]",
            "unpack = rbx.geo.cli:unpack [geo]",
        ],
    },
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "": ["*.md", "*.yaml"],
    },
)



import csv
from io import StringIO
import logging
from os import path
import shelve

import arrow

from ..exceptions import ClientException
from . import Client, HttpAuth, retry

logger = logging.getLogger(__name__)


class AdsquareClient(Client):
    """Client for adsquare API.

    The API documentation can be found on the Platform wiki:
    https://github.com/rockabox/rbx/wiki/adsquare

    """

    AUTH_PATH = "auth/login"
    ENDPOINT = "https://amp.adsquare.com/api/v1/"
    STANDARD_COMPANY_ID = "588b6428e4b0d7d5f6895bd1"
    TIME_FORMAT = "YYYY-MM-DDTHH:mm:ss.SSS[Z]"

    def __init__(self, company_id, dsp_id, destination="/tmp"):
        super().__init__()
        self.company_id = company_id
        self.dsp_id = dsp_id
        self.shelf = path.join(destination, "adsquare.segments")

    @property
    def auth(self):
        """adsquare uses Digest Authentication."""
        return HttpAuth(self.token, key="X-AUTH-TOKEN")

    @retry.Retry(deadline=300.0)
    def create_config(self, country_code, name, start_date, end_date):
        """Create a new observation configuration.

        The start and end date are expected to be in ISO 8601, though any format understood by
        Arrow will work.
        """
        path = f"measurement/configs/byCompany/{self.company_id}"
        response = self.request(
            "post",
            path,
            data={
                "dspId": self.dsp_id,
                "countryCode": country_code,
                "name": name,
                "active": True,
                "observationMode": "full",
                "radiusInMeter": 50,
                "validFrom": arrow.get(start_date).format(self.TIME_FORMAT),
                "validTo": arrow.get(end_date).format(self.TIME_FORMAT),
            },
        )
        return response["id"]

    def disable_webhook(self, config_id):
        """Disable the registered webhook for the given configuration ID."""
        self.register_webhook(config_id, disable=True)

    @retry.Retry(deadline=300.0)
    def download_config(self, config_id):
        """Download the list of site locations uploaded to the given configuration.

        The points are returned as a dictionary of (lat, long) tuples keyed by assigned ID.
        """
        data = self.request(
            "get",
            f"measurement/configs/{config_id}/dataset/poi",
            content_type="text/plain",
        )

        pois = {}
        reader = csv.DictReader(StringIO(data))
        for row in reader:
            pois[row["id"]] = (row["lat"], row["lon"])

        return pois

    def get_segments(self, company_ids=None, country_code=None):
        """Get segments from local cache."""
        with shelve.open(self.shelf) as shelf:
            try:
                segments = shelf["segments"]
            except KeyError:
                # The cache is empty
                segments = []

        # Always include Standard segments (i.e.: Company ID = adsquare).
        company_ids = company_ids or []
        company_ids.append(self.STANDARD_COMPANY_ID)

        # Exclude company_ids that are either None or empty.
        company_ids[:] = [cid for cid in company_ids if cid]

        segment_list = []
        for segment in segments:
            if segment["company_id"] not in company_ids:
                continue
            if country_code and country_code not in [segment["country"], "ZZ"]:
                continue
            segment_list.append(segment)

        return segment_list

    @retry.Retry(deadline=300.0)
    def list_segments(self):
        """List all segments available.

        According to the documentation, there is no pagination on this endpoint, so we expect the
        full list to be returned with a single call.

        See: https://docs.adsquare.com/#!/enrichmentMetaApi/getEnrichmentAudiences_1
        """
        path = f"enrichmentMeta/audiences/{self.dsp_id}"
        response = self.request("get", path, data={"namev2": "true"})

        return [
            {
                "company_id": audience["companyId"],
                "country": audience["countryCode"],
                "cpm": audience["cpm"],
                "currency": audience["currency"],
                "id": audience["audienceId"],
                "name": audience["name"],
                "owner": "Standard"
                if audience["companyId"] == self.STANDARD_COMPANY_ID
                else "Custom",
                "type": audience["type"],
            }
            for audience in response["audiences"]
        ]

    def pause_config(self, config_id):
        """Pause the given configuration."""
        self.update_config(config_id=config_id, active=False)

    @retry.Retry(deadline=300.0)
    def register_webhook(self, config_id, disable=False, feedback_url=None):
        """Register a Feedback URL webhook with adsquare for the given configuration ID.

        The same method can be used to re-enable an existing/disabled webhook.
        """
        data = {
            "enabled": not disable,
            "method": "post",
            "contentType": "csv",
        }
        if feedback_url:
            data["feedbackURL"] = feedback_url

        if not disable and not feedback_url:
            raise ClientException("Cannot register a webhook without a URL")

        return self.request(
            "post", f"measurement/configs/{config_id}/feedback", data=data
        )

    def resume_config(self, config_id):
        """Resume the given configuration."""
        self.update_config(config_id=config_id, active=True)

    def store_segments(self):
        """Retrieve the list of segments, and then cache the results on a shelf.

        As no caching mechanism exists on the API, we simply override the content of the cache
        with the results from the list_segments() call.
        """
        segments = self.list_segments()
        with shelve.open(self.shelf) as shelf:
            shelf["segments"] = segments

    @retry.Retry(deadline=300.0)
    def update_config(self, config_id, active=None, start_date=None, end_date=None):
        """Update the given configuration ID.

        Only active and start/end dates can be changed.
        """
        data = {}

        if active is not None:
            data["active"] = active

        if start_date:
            data["validFrom"] = arrow.get(start_date).format(self.TIME_FORMAT)

        if end_date:
            data["validTo"] = arrow.get(end_date).format(self.TIME_FORMAT)

        return self.request("post", f"measurement/configs/{config_id}", data=data)

    @retry.Retry(deadline=300.0)
    def upload_config(self, config_id, points):
        """Upload a list of site locations to the given configuration.

        The points is expected to be a list of (lat, long) tuples as floating values.
        """
        data = ["id,lon,lat"]
        for i, point in enumerate(points):
            data.append(f"{i},{point[1]},{point[0]}")

        # Make sure the payload has a trailing newline
        data.append("")

        self.request(
            "put", f"measurement/configs/{config_id}/dataset/poi", data="\n".join(data)
        )



from datetime import timedelta
from functools import wraps
import logging
import threading
import traceback
import warnings

from redis import StrictRedis
from rq import Queue as RQ, SimpleWorker, get_current_job
from rq.job import cancel_job
from rq.registry import ScheduledJobRegistry

from .scheduler import Scheduler

connection = threading.local()
logger = logging.getLogger(__name__)


class Queue:
    """The Queue provides an interface to farm out jobs to be processed asynchronously."""

    def __init__(
        self,
        host="0.0.0.0",
        port=6379,
        db=0,
        module=None,
        conn=None,
        is_async=True,
        no_delay=False,
    ):
        """Initialise the Job Queue.

        Parameters:
            host (str):
                The Redis cluster host.
                Uses '0.0.0.0' by default.
            port (int):
                The Redis cluster port.
                Uses port 6379 by default.
            db (int):
                The Redis cluster db.
                Uses 0 by default.
            module:
                The default Queue uses either a method, or as string reference to an importable
                method.
                For string references, this paramater can be used to always load the job method
                from a specific module.

                >>> queue = Queue(module='src.tasks')
                >>> queue.enqueue('method')
                Job(function='src.tasks.method', ...)

            conn (StrictRedis):
                A redis connection may be provided instead of the host/port/db combination.
            is_async (bool):
                Set to True to skip the workers and perform the job instantly.
            no_delay (bool):
                Use this flag to cancel the scheduled/delaying feature, and enqueue a delayed job
                immediately. Useful for testing purposes.
        """
        if conn:
            connection.redis = conn
        else:
            connection.redis = StrictRedis(host=host, port=port, db=db)
        self.connection = connection.redis

        self.queue = RQ(connection=self.connection, is_async=is_async)
        self.module = module
        self.no_delay = no_delay

    @property
    def count(self):
        return self.queue.count

    @property
    def is_async(self):
        return self.queue.is_async

    @property
    def name(self):
        return self.queue.name

    @property
    def exception_handlers(self):
        """The handlers must be listed in order of expected execution."""
        return [self._retry_on_failure, self._requeue_job, self._move_to_failed_queue]

    @property
    def failed_queue(self):
        return self.queue.failed_job_registry

    def _move_to_failed_queue(self, job, *exc_info):
        """This handler moves the job to the failed queue."""
        exc_string = "".join(traceback.format_exception(*exc_info))
        logger.debug("Moving job to failed queue")
        self.failed_queue.add(job, ttl=job.failure_ttl, exc_string=exc_string)

        logger.debug(
            "Job {} failed <{}>.".format(job.id, job.func_name),
            exc_info=exc_info,
            extra={"arguments": job.args},
        )

    def _requeue_job(self, job, *exc_info):
        """This handler does the requeuing, according to the 'failures' meta data set by the
        previous handler.

        If the retry is given a sequence of exception types, the method will only be retried if the
        exception raised is of one of these types.

        When the maximum retry attempts has been reached, we make sure to clear the 'failures'
        flag, and move the job to the next handler in the stack (which will move it to the failed
        queue).
        """
        failures = job.meta.get("failures", 0)
        retries = job.meta.get("retries", 3)
        if "retry" not in job.meta or failures >= retries:
            job.meta.pop("failures", None)
            job.save_meta()
            return True

        exceptions = job.meta.get("exceptions")
        if exceptions and not isinstance(exc_info[1], exceptions):
            logger.debug("Un-retryable exception")
            return True

        logger.debug("Requeuing {!r}".format(job))
        self.requeue_job(job)

        return False

    def _retry_on_failure(self, job, *exc_info):
        """This handler flags jobs which are retry-able ('retry' = True) by incrementing the
        'failures' meta data.

        It will then pass on the handling of the requeuing itself to the next handler in the stack.
        If we were to requeue the job at this stage, what would happen is the job would be queued
        up before the next handler had the chance to fallback to the next, ending in an infinite
        loop of failing jobs with the same ID.

        The retry() decorator is provided to flag any method as retry-able.

        Also note that scheduled jobs do never get retried, and skip the failed queue entirely.
        They will get retried at the next run anyway.
        """
        if "scheduled" in job.meta:
            # This is a scheduled job failing. Skip the error handling.
            return False

        job.meta.setdefault("failures", 0)
        logger.debug("META: {}".format(job.meta))

        if "retry" in job.meta:
            job.meta["failures"] += 1
            logger.debug("Failures: {}".format(job.meta["failures"]))

        job.save_meta()

        return True

    def as_sync(self):
        """Return a Queue instance with the is_async flag set to False."""
        return Queue(
            conn=self.connection,
            is_async=False,
            module=self.module,
            no_delay=self.no_delay,
        )

    def cancel_job(self, job):
        """Cancel the given job."""
        cancel_job(job.id, connection=self.connection)
        job.delete()

    def cancel_jobs(self, job_ids):
        """Cancel of bunch of jobs given their IDs."""
        for job_id in job_ids:
            self.cancel_job(self.fetch_job(job_id))

    def empty(self):
        """Empty all queues."""
        self.queue.empty()

        for job_id in self.failed_queue.get_job_ids():
            job = self.queue.fetch_job(job_id=job_id)
            self.failed_queue.remove(job)

        self.cancel_jobs([job.id for job in self.get_scheduled_jobs()])

    def enqueue(self, function, *args, enqueue_in=None, **kwargs):
        """Register a job to be processed.

        Parameters:
            function (str|function):
                A function, or a string reference to a function. If self.module is set, it is used
                as the default location for looking up method names.
            enqueue_in (int):
                By default, jobs are scheduled to be executed as soon as a worker is available. The
                execution can be further delayed by specifying how to long to wait. The value is
                expressed in seconds.
            args (tuple):
                The positional arguments to pass on to the function.
            kwargs (dict):
                The keyword arguments to pass on to the function.

        Returns:
            The job to be executed.
        """
        if self.module and isinstance(function, str):
            function = f"{self.module}.{function}"

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning, module="rq")

            if (
                enqueue_in
                and isinstance(enqueue_in, int)
                and self.is_async
                and not self.no_delay
            ):
                logger.debug(
                    f"Job scheduled in {enqueue_in}s: {function}, {args}, {kwargs}"
                )
                return self.queue.enqueue_in(
                    timedelta(seconds=enqueue_in), function, *args, **kwargs
                )
            else:
                logger.debug(f"Job queued: {function}, {args}, {kwargs}")
                return self.queue.enqueue(function, *args, **kwargs)

    def fetch_job(self, job_id):
        """Fetch a job for a Job ID."""
        return self.queue.fetch_job(job_id)

    def get_jobs(self, *args, **kwargs):
        """Get all jobs."""
        return self.queue.get_jobs(*args, **kwargs)

    def get_scheduled_jobs(self):
        """Get all scheduled jobs."""
        registry = ScheduledJobRegistry(queue=self.queue)
        return [self.fetch_job(job_id=job_id) for job_id in registry.get_job_ids()]

    def get_worker(self, *args, **kwargs):
        """Return an instance of a worker."""

        class Worker(SimpleWorker):
            def work(self, *args, **kwargs):
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore", category=DeprecationWarning, module="rq"
                    )
                    super().work(*args, **kwargs)

        return Worker(
            [self.queue],
            *args,
            connection=self.connection,
            disable_default_exception_handler=True,
            exception_handlers=self.exception_handlers,
            **kwargs,
        )

    def requeue_failed_job(self, job_id):
        """Requeue a failed job to its original queue."""
        logger.debug("Requeuing job {}".format(job_id))
        self.failed_queue.requeue(job_id)

    def requeue_failed_jobs(self, job_ids):
        """Requeue of bunch of failed jobs given their IDs."""
        for job_id in job_ids:
            self.requeue_failed_job(job_id)

    def requeue_job(self, job):
        """Requeue a job that is due to be re-tried after a failure.

        These jobs are pushed back into the queue without having made it to the failed queue.

        The requeue_job method provided by the Queue class does not keep the meta data, which we
        use to check how often the job has been tried. We therefore requeue these jobs by enqueing
        the original signature, including the meta data and the original Job ID.
        """
        self.queue.enqueue(
            job.func, args=job.args, kwargs=job.kwargs, job_id=job.id, meta=job.meta
        )

    def with_no_delay(self):
        """Return a Queue instance with the no_delay flag set to True."""
        return Queue(
            conn=self.connection,
            is_async=self.is_async,
            module=self.module,
            no_delay=True,
        )


def retry(exceptions=None, max_retries=3):
    """Decorator used to move a failing job back into the queue.

    Parameters:
        exceptions (Exception):
            An exception type, or a tuple of types, used to determine whether, when raised from
            within a job, the method should be considered retryable.

            e.g.:
            >>> retry(exceptions=ValueError)
            >>> retry(exceptions=(TransientError, ValueError))

        max_retries (int):
            The number of attempts before giving up. Defaults to 3.
    """

    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            job = get_current_job(connection=connection.redis)
            job.meta["retry"] = True
            job.meta["retries"] = max_retries
            job.meta["exceptions"] = exceptions
            job.save()

            return function(*args, **kwargs)

        return wrapper

    return decorator


__all__ = ["get_current_job", "Queue", "retry", "Scheduler"]
